Attribute VB_Name = "Module1"
Option Explicit

Public Const Pi As Double = 3.14159265359

Private Function rndGauss(ByVal Mu As Double, ByVal Sigma As Double, ByVal Bounds As Double) As Double

Dim x As Double
Dim y As Double
Dim Density As Double
 
Do
    x = 2 * Bounds * Rnd + (Mu - Bounds)
    y = Rnd
    Density = (Sqr(2 * 3.1415 * Sigma) ^ (-1)) * Exp(-((x - Mu) ^ 2) / (2 * Sigma))
Loop Until y < Density
 
rndGauss = x

End Function

Private Function rndRange(ByVal Min As Double, ByVal Max As Double) As Double

Dim Range As Double
Dim Frac As Double

Range = Abs(Min) + Abs(Max)

Frac = Rnd * Range

rndRange = Frac + Min

End Function

Private Sub QuickSort(ByRef vSort() As Double, Optional ByVal Index As Long = 1, Optional ByVal lngStart As Variant, Optional ByVal lngEnd As Variant)

If IsMissing(lngStart) Then lngStart = LBound(vSort)
If IsMissing(lngEnd) Then lngEnd = UBound(vSort)

Dim i As Long
Dim j As Long
Dim h As Double
Dim x As Double
Dim u As Long
Dim lb_dim As Long
Dim ub_dim As Long

lb_dim = LBound(vSort, 2)
ub_dim = UBound(vSort, 2)

i = lngStart
j = lngEnd
x = vSort((lngStart + lngEnd) / 2, Index)

Do
    While (vSort(i, Index) < x): i = i + 1: Wend
    While (vSort(j, Index) > x): j = j - 1: Wend

    If (i <= j) Then
        For u = lb_dim To ub_dim
            h = vSort(i, u)
            vSort(i, u) = vSort(j, u)
            vSort(j, u) = h
        Next
        i = i + 1
        j = j - 1
    End If
Loop Until (i > j)

If (lngStart < j) Then QuickSort vSort, Index, lngStart, j
If (i < lngEnd) Then QuickSort vSort, Index, i, lngEnd

End Sub

Private Function Resample(ByVal Num As Long) As Long()

Dim i As Long
Dim j As Long
Dim Candidate As Long
Dim IsValid As Boolean
Dim OutVec() As Long
ReDim OutVec(1 To Num) As Long

For i = 1 To Num
    IsValid = False
    Do Until IsValid
        Candidate = CLng(rndRange(1, Num))
        For j = 1 To i
            If OutVec(j) = Candidate Then Exit For
        Next
        If j >= i Then
            OutVec(i) = Candidate
            IsValid = True
        End If
    Loop
Next

Resample = OutVec

End Function

Public Function optimGA(ByVal Fun As String, ByVal nVars As Long, ByRef Domain() As Double, Optional PopSize As Long = 100, Optional MaxGenerations As Long = 1000, Optional MutationProbability As Double = 0.7, Optional RecombineProbability As Double = 1, Optional Eps As Double = 0.00001, Optional Minimize As Boolean = True) As Double()

Dim Max As Boolean
Dim Population() As Double
Dim i As Long
Dim j As Long
Dim k As Long
Dim CurrVars() As Double
Dim EpsEmp As Double
Dim retVal() As Double
Dim RecombinePair1() As Long
Dim RecombinePair2() As Long
Dim Temp() As Double
Dim Mutate As Double
Dim Recombine As Double

Max = Not Minimize

ReDim Population(1 To PopSize, 1 To nVars + 1)
ReDim retVal(1 To nVars)

For i = 1 To PopSize
    For j = 1 To nVars
        Population(i, j) = rndRange(Domain(0), Domain(1))
    Next
Next

For i = 1 To MaxGenerations
    Debug.Print "Generation " & i
        
    For j = 1 To PopSize
        ReDim CurrVars(0 To nVars - 1)
        For k = 0 To nVars - 1
            CurrVars(k) = Population(j, k + 1)
        Next
        Population(j, nVars + 1) = CallByName(Form1, Fun, VbMethod, CurrVars)
    Next
    
    Call QuickSort(Population, nVars + 1) 'Bislang nur aufsteigende Sortierung -> Nur Minimierung mglich.
    
    For j = (PopSize / 2) + 1 To PopSize
        For k = 1 To nVars
            Population(j, k) = 0
        Next
    Next
    
    EpsEmp = Abs(Population(1, nVars + 1) - Population(PopSize / 2, nVars + 1))
    If EpsEmp < Eps Then
        For j = 1 To nVars
            retVal(j) = Population(1, j)
        Next
        optimGA = retVal
        Exit Function
    End If
    
    RecombinePair1 = Resample(PopSize / 2)
    RecombinePair2 = Resample(PopSize / 2)
    
    Temp = Population
    For j = 1 To PopSize / 2
        Mutate = Rnd
        If Mutate >= MutationProbability Then Mutate = 1 Else Mutate = 0
        
        Recombine = Rnd
        If Recombine >= (1 - RecombineProbability) Then
            For k = 1 To nVars
                Temp(j + (PopSize / 2), k) = ((Population(RecombinePair1(j), k) + Population(RecombinePair2(j), k)) / 2) + (Mutate * rndGauss(0, Domain(1) - Domain(0), 5))
                If Temp(j + (PopSize / 2), k) < Domain(0) Then Temp(j + (PopSize / 2), k) = Domain(0)
                If Temp(j + (PopSize / 2), k) > Domain(1) Then Temp(j + (PopSize / 2), k) = Domain(1)
            Next
        Else
                Temp(j + (PopSize / 2), k) = Population(j, k) + (Mutate * rndGauss(0, Domain(1) - Domain(0), 5))
                If Temp(j + (PopSize / 2), k) < Domain(0) Then Temp(j + (PopSize / 2), k) = Domain(0)
                If Temp(j + (PopSize / 2), k) > Domain(1) Then Temp(j + (PopSize / 2), k) = Domain(1)
        End If
    Next
    Population = Temp
Next

optimGA = retVal

End Function
